@@ -52,6 +52,10 @@
 #include <stdarg.h>
 #include <conio.h>
 
+#ifdef DJGPP
+#include <pc.h>
+#endif
+
 // Ah the wonders of TRUE and FALSE, heaven help us if they`re enums
 #ifndef FALSE
 #define FALSE 0
@@ -727,19 +731,42 @@ void tkQuit(void)
 
 /******************************************************************************/
 
+#if defined( __WATCOMC__)
 void outpb(unsigned char byte);
 #pragma aux outpb = \
 	"mov dx,03c9h" \
 	"out dx,al" \
 	parm [ax] \
 	modify exact [dx];
+#endif
 
-void set_onecolor(int index,int R,int G,int B)
+static unsigned char Rr,Gg,Bb,Cindex;
+void  set_onecolor(int cindex,int R,int G,int B)
 {
-	outp(0x3c8,index);
+#if defined( __WATCOMC__ )
+	outp(0x3c8,cindex);
 	outpb(R);
 	outpb(G);
 	outpb(B);
+#endif
+
+#if defined( DJGPP )
+	Rr=R;Gg=G;Bb=B;
+	Cindex = cindex;
+
+	asm ("
+		movw	$0x3c8,%dx
+		movw	_Cindex,%ax
+		outb	%al,%dx
+		inc	%dx
+		movb	_Rr,%al
+		outb	%al,%dx
+		movb	_Gg,%al
+		outb	%al,%dx
+		movb	_Bb,%al
+		outb	%al,%dx
+	");
+#endif
 }
 
 void tkSetOneColor(int index, float r, float g, float b)
