@@ -1,9 +1,9 @@
-/* $Id: image.c,v 1.6 1996/11/13 03:58:31 brianp Exp $ */
+/* $Id: image.c,v 1.9 1997/02/09 20:05:03 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  2.0
- * Copyright (C) 1995-1996  Brian Paul
+ * Version:  2.2
+ * Copyright (C) 1995-1997  Brian Paul
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +23,15 @@
 
 /*
  * $Log: image.c,v $
+ * Revision 1.9  1997/02/09 20:05:03  brianp
+ * new arguments for gl_pixel_addr_in_image()
+ *
+ * Revision 1.8  1997/02/09 18:52:53  brianp
+ * added GL_EXT_texture3D support
+ *
+ * Revision 1.7  1997/01/09 21:25:54  brianp
+ * initialize image reference count to zero
+ *
  * Revision 1.6  1996/11/13 03:58:31  brianp
  * fixed undefined "format" variable in gl_unpack_image()
  *
@@ -171,30 +180,33 @@ GLint gl_components_in_format( GLenum format )
 
 
 /*
- * Return the address of a pixel in an image.  Pixel unpacking/packing
- * parameters are observed according to 'packing'.
+ * Return the address of a pixel in an image (actually a volume).
+ * Pixel unpacking/packing parameters are observed according to 'packing'.
  * Input:  image - start of image data
  *         width, height - size of image
  *         format - image format
  *         type - pixel component type
  *         packing - GL_TRUE = use packing params
  *                   GL_FALSE = use unpacking params.
- *         row, column - location of pixel whose address is to be returned
- * Return:  address of pixel at (row,column) in image or NULL if error.
+ *         img - which image in the volume (0 for 2-D images)
+ *         row, column - location of pixel in the image
+ * Return:  address of pixel at (image,row,column) in image or NULL if error.
  */
-GLvoid *gl_pixel_addr_in_image( GLcontext *ctx,
+GLvoid *gl_pixel_addr_in_image( struct gl_pixelstore_attrib *packing,
                                 const GLvoid *image, GLsizei width,
                                 GLsizei height, GLenum format, GLenum type,
-                                GLboolean packing,
-                                GLint row, GLint column )
+                                GLint img, GLint row, GLint column )
 {
    GLint bytes_per_comp;   /* bytes per component */
    GLint comp_per_pixel;   /* components per pixel */
    GLint comps_per_row;    /* components per row */
    GLint pixels_per_row;   /* pixels per row */
+   GLint bytes_per_image;
+   GLint rows_per_image;
    GLint alignment;        /* 1, 2 or 4 */
    GLint skiprows;
    GLint skippixels;
+   GLint skipimages;       /* for 3-D */
    GLubyte *pixel_addr;
 
    /* Compute bytes per component */
@@ -209,30 +221,22 @@ GLvoid *gl_pixel_addr_in_image( GLcontext *ctx,
       return NULL;
    }
 
-   if (packing) {
-      /* Use PACKING parameters */
-      alignment = ctx->Pack.Alignment;
-      if (ctx->Pack.RowLength>0) {
-         pixels_per_row = ctx->Pack.RowLength;
-      }
-      else {
-         pixels_per_row = width;
-      }
-      skiprows = ctx->Pack.SkipRows;
-      skippixels = ctx->Pack.SkipPixels;
+   alignment = packing->Alignment;
+   if (packing->RowLength>0) {
+      pixels_per_row = packing->RowLength;
    }
    else {
-      /* Use UNPACKING parameters */
-      alignment = ctx->Unpack.Alignment;
-      if (ctx->Unpack.RowLength>0) {
-         pixels_per_row = ctx->Unpack.RowLength;
-      }
-      else {
-         pixels_per_row = width;
-      }
-      skiprows = ctx->Unpack.SkipRows;
-      skippixels = ctx->Unpack.SkipPixels;
+      pixels_per_row = width;
    }
+   if (packing->ImageHeight>0) {
+      rows_per_image = packing->ImageHeight;
+   }
+   else {
+      rows_per_image = height;
+   }
+   skiprows = packing->SkipRows;
+   skippixels = packing->SkipPixels;
+   skipimages = packing->SkipImages;
 
    if (type==GL_BITMAP) {
       /* BITMAP data */
@@ -241,7 +245,10 @@ GLvoid *gl_pixel_addr_in_image( GLcontext *ctx,
       bytes_per_row = alignment
                     * CEILING( comp_per_pixel*pixels_per_row, 8*alignment );
 
+      bytes_per_image = bytes_per_row * rows_per_image;
+
       pixel_addr = (GLubyte *) image
+                 + (skipimages + img) * bytes_per_image
                  + (skiprows + row) * bytes_per_row
                  + (skippixels + column) / 8;
    }
@@ -259,10 +266,13 @@ GLvoid *gl_pixel_addr_in_image( GLcontext *ctx,
                        * CEILING( bytes_per_row, alignment );
       }
 
+      bytes_per_image = bytes_per_comp * comps_per_row * rows_per_image;
+
       /* Copy/unpack pixel data to buffer */
       pixel_addr = (GLubyte *) image
-                 + (skiprows + row) * comps_per_row * bytes_per_comp
-                 + (skippixels + column) * comp_per_pixel * bytes_per_comp;
+                 + (skipimages + img) * bytes_per_image
+                 + (skiprows + row) * bytes_per_comp * comps_per_row
+                 + (skippixels + column) * bytes_per_comp * comp_per_pixel;
    }
 
    return (GLvoid *) pixel_addr;
@@ -292,6 +302,35 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
                                   GLenum destType,
                                   const GLvoid *pixels,
                                   GLboolean interleave )
+{ 
+   return gl_unpack_image3D(ctx, width, height, 1,
+                            srcFormat, srcType, destType, pixels, interleave);
+}
+
+
+
+/* 
+ * Unpack a 2-D/3-D image from user-supplied address, returning a pointer to
+ * a new gl_image struct.
+ * This function is always called by a higher-level unpack function such
+ * as gl_unpack_texsubimage() or gl_unpack_bitmap().
+ *
+ * Input:  width, height, depth - size in pixels
+ *         srcFormat - format of incoming pixel data, ignored
+ *                      if srcType and destType is BITMAP.
+ *         srcType - GL_UNSIGNED_BYTE .. GL_FLOAT
+ *         destType - store image as GL_UNSIGNED_BYTE, GL_FLOAT, or GL_BITMAP.
+ *                    if GL_UNSIGNED_BYTE, srctype must be GL_UNSIGNED_BYTE.
+ *                    if GL_BITMAP, srctype must be GL_BITMAP.
+ *         interleave - if TRUE, srctype and destType must be GL_UNSIGNED_BYTE
+ *         pixels - pointer to unpacked image.
+ */
+struct gl_image *gl_unpack_image3D( GLcontext *ctx,
+                                    GLint width, GLint height, GLint depth,
+                                    GLenum srcFormat, GLenum srcType,
+                                    GLenum destType,
+                                    const GLvoid *pixels,
+                                    GLboolean interleave )
 {
    GLint components;
 
@@ -299,14 +338,13 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
 
    if (srcType==GL_BITMAP || destType==GL_BITMAP) {
       struct gl_image *image;
-      GLint bytes, i, width_in_bytes;
+      GLint bytes, i, width_in_bytes, d;
       GLubyte *buffer, *dst;
-
       assert( srcType==GL_BITMAP );
       assert( destType==GL_BITMAP );
 
       /* Alloc dest storage */
-      bytes = (width+7)/8 * height;
+      bytes = ((width+7)/8 * height) * depth;
       if (bytes>0 && pixels!=NULL) {
          buffer = (GLubyte *) malloc( bytes );
          if (!buffer) {
@@ -315,16 +353,19 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
          /* Copy/unpack pixel data to buffer */
          width_in_bytes = CEILING( width, 8 );
          dst = buffer;
-         for (i=0;i<height;i++) {
-            GLvoid *src = gl_pixel_addr_in_image( ctx, pixels, width, height,
-                                                  GL_COLOR_INDEX, srcType,
-                                                  GL_FALSE, i, 0 );
-            if (!src) {
-               free(buffer);
-               return NULL;
+         for (d=0; d<depth; d++) {
+            for (i=0; i<height; i++) {
+               GLvoid *src = gl_pixel_addr_in_image( &ctx->Unpack, pixels,
+                                                     width, height,
+                                                     GL_COLOR_INDEX, srcType,
+                                                     d, i, 0 );
+               if (!src) {
+                  free(buffer);
+                  return NULL;
+               }
+               MEMCPY( dst, src, width_in_bytes );
+               dst += width_in_bytes;
             }
-            MEMCPY( dst, src, width_in_bytes );
-            dst += width_in_bytes;
          }
          /* Bit flipping */
          if (ctx->Unpack.LsbFirst) {
@@ -340,11 +381,13 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
       if (image) {
          image->Width = width;
          image->Height = height;
+         image->Depth = depth;
          image->Components = 0;
          image->Format = GL_COLOR_INDEX;
          image->Type = GL_BITMAP;
          image->Interleaved = GL_FALSE;
          image->Data = buffer;
+         image->RefCount = 0;
       }
       else {
          free( buffer );
@@ -364,47 +407,48 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
       struct gl_image *image;
       GLint width_in_bytes;
       GLubyte *buffer, *dst;
-      GLint i;
-
+      GLint i, d;
       assert( srcType==GL_UNSIGNED_BYTE );
 
       width_in_bytes = width * components * sizeof(GLubyte);
-      buffer = malloc( height * width_in_bytes );
+      buffer = malloc( height * width_in_bytes * depth );
       if (!buffer) {
          return NULL;
       }
-
       /* Copy/unpack pixel data to buffer */
       dst = buffer;
-      for (i=0;i<height;i++) {
-         GLubyte *src = (GLubyte *) gl_pixel_addr_in_image( ctx, pixels, width,
-                                  height, srcFormat, srcType, GL_FALSE, i, 0 );
-         if (!src) {
-            free(buffer);
-            return NULL;
-         }
-         if (interleave) {
-            GLint j, k;
-            for (j=0;j<width;j++) {
-               for (k=0;k<components;k++) {
-                  dst[k*width+j] = src[j*components+k];
+      for (d=0; d<depth; d++ ) {
+         for (i=0;i<height;i++) {
+            GLubyte *src = (GLubyte *) gl_pixel_addr_in_image( &ctx->Unpack,
+                           pixels, width, height, srcFormat, srcType, d, i, 0 );
+            if (!src) {
+               free(buffer);
+               return NULL;
+            }
+            if (interleave) {
+               GLint j, k;
+               for (j=0;j<width;j++) {
+                  for (k=0;k<components;k++) {
+                     dst[k*width+j] = src[j*components+k];
+                  }
                }
             }
+            else {
+               MEMCPY( dst, src, width_in_bytes );
+            }
+            dst += width_in_bytes;
          }
-         else {
-            MEMCPY( dst, src, width_in_bytes );
-         }
-         dst += width_in_bytes;
       }
 
       if (ctx->Unpack.LsbFirst) {
-         gl_flip_bytes( buffer, height * width_in_bytes );
+         gl_flip_bytes( buffer, height * width_in_bytes * depth );
       }
 
       image = (struct gl_image *) malloc( sizeof(struct gl_image) );
       if (image) {
          image->Width = width;
          image->Height = height;
+         image->Depth = depth; 
          image->Components = components;
          image->Format = srcFormat;
          image->Type = GL_UNSIGNED_BYTE;
@@ -421,125 +465,129 @@ struct gl_image *gl_unpack_image( GLcontext *ctx,
       struct gl_image *image;
       GLfloat *buffer, *dst;
       GLint elems_per_row;
-      GLint i, j;
-
+      GLint i, j, d;
       elems_per_row = width * components;
-      buffer = (GLfloat *) malloc( height * elems_per_row * sizeof(GLfloat) );
+      buffer = (GLfloat *) malloc( height * elems_per_row * sizeof(GLfloat) * depth);
       if (!buffer) {
          return NULL;
       }
 
       dst = buffer;
-      for (i=0;i<height;i++) {
-         GLvoid *src = gl_pixel_addr_in_image( ctx, pixels, width, height,
-                                               srcFormat, srcType, GL_FALSE,
-                                               i, 0 );
-         if (!src) {
-            free(buffer);
-            return NULL;
-         }
+      /**      img_pixels= pixels;*/
+      for (d=0; d<depth; d++) {
+         for (i=0;i<height;i++) {
+            GLvoid *src = gl_pixel_addr_in_image( &ctx->Unpack, pixels,
+                                                  width, height,
+                                                  srcFormat, srcType,
+                                                  d, i, 0 );
+            if (!src) {
+               free(buffer);
+               return NULL;
+            }
 
-         switch (srcType) {
-            case GL_UNSIGNED_BYTE:
-               for (j=0;j<elems_per_row;j++) {
-                  *dst++ = (GLfloat) ((GLubyte*)src)[j];
-               }
-               break;
-            case GL_BYTE:
-               for (j=0;j<elems_per_row;j++) {
-                  *dst++ = (GLfloat) ((GLbyte*)src)[j];
-               }
-               break;
-            case GL_UNSIGNED_SHORT:
-               if (ctx->Unpack.SwapBytes) {
+            switch (srcType) {
+               case GL_UNSIGNED_BYTE:
                   for (j=0;j<elems_per_row;j++) {
-                     GLushort value = ((GLushort*)src)[j];
-                     value = ((value >> 8) & 0xff) | ((value&0xff) << 8);
-                     *dst++ = (GLfloat) value;
+                     *dst++ = (GLfloat) ((GLubyte*)src)[j];
                   }
-               }
-               else {
+                  break;
+               case GL_BYTE:
                   for (j=0;j<elems_per_row;j++) {
-                     *dst++ = (GLfloat) ((GLushort*)src)[j];
+                     *dst++ = (GLfloat) ((GLbyte*)src)[j];
                   }
-               }
-               break;
-            case GL_SHORT:
-               if (ctx->Unpack.SwapBytes) {
-                  for (j=0;j<elems_per_row;j++) {
-                     GLshort value = ((GLshort*)src)[j];
-                     value = ((value >> 8) & 0xff) | ((value&0xff) << 8);
-                     *dst++ = (GLfloat) value;
+                  break;
+               case GL_UNSIGNED_SHORT:
+                  if (ctx->Unpack.SwapBytes) {
+                     for (j=0;j<elems_per_row;j++) {
+                        GLushort value = ((GLushort*)src)[j];
+                        value = ((value >> 8) & 0xff) | ((value&0xff) << 8);
+                        *dst++ = (GLfloat) value;
+                     }
                   }
-               }
-               else {
-                  for (j=0;j<elems_per_row;j++) {
-                     *dst++ = (GLfloat) ((GLshort*)src)[j];
+                  else {
+                     for (j=0;j<elems_per_row;j++) {
+                        *dst++ = (GLfloat) ((GLushort*)src)[j];
+                     }
                   }
-               }
-               break;
-            case GL_UNSIGNED_INT:
-               if (ctx->Unpack.SwapBytes) {
-                  GLuint value;
-                  for (j=0;j<elems_per_row;j++) {
-                     value = ((GLuint*)src)[j];
-                     value = ((value & 0xff000000) >> 24)
-                           | ((value & 0x00ff0000) >> 8)
-                           | ((value & 0x0000ff00) << 8)
-                           | ((value & 0x000000ff) << 24);
-                     *dst++ = (GLfloat) value;
+                  break;
+               case GL_SHORT:
+                  if (ctx->Unpack.SwapBytes) {
+                     for (j=0;j<elems_per_row;j++) {
+                        GLshort value = ((GLshort*)src)[j];
+                        value = ((value >> 8) & 0xff) | ((value&0xff) << 8);
+                        *dst++ = (GLfloat) value;
+                     }
                   }
-               }
-               else {
-                  for (j=0;j<elems_per_row;j++) {
-                     *dst++ = (GLfloat) ((GLuint*)src)[j];
+                  else {
+                     for (j=0;j<elems_per_row;j++) {
+                        *dst++ = (GLfloat) ((GLshort*)src)[j];
+                     }
                   }
-               }
-               break;
-            case GL_INT:
-               if (ctx->Unpack.SwapBytes) {
-                  GLint value;
-                  for (j=0;j<elems_per_row;j++) {
-                     value = ((GLint*)src)[j];
-                     value = ((value & 0xff000000) >> 24)
-                           | ((value & 0x00ff0000) >> 8)
-                           | ((value & 0x0000ff00) << 8)
-                           | ((value & 0x000000ff) << 24);
-                     *dst++ = (GLfloat) value;
+                  break;
+               case GL_UNSIGNED_INT:
+                  if (ctx->Unpack.SwapBytes) {
+                     GLuint value;
+                     for (j=0;j<elems_per_row;j++) {
+                        value = ((GLuint*)src)[j];
+                        value = ((value & 0xff000000) >> 24)
+                              | ((value & 0x00ff0000) >> 8)
+                              | ((value & 0x0000ff00) << 8)
+                              | ((value & 0x000000ff) << 24);
+                        *dst++ = (GLfloat) value;
+                     }
                   }
-               }
-               else {
-                  for (j=0;j<elems_per_row;j++) {
-                     *dst++ = (GLfloat) ((GLint*)src)[j];
+                  else {
+                     for (j=0;j<elems_per_row;j++) {
+                        *dst++ = (GLfloat) ((GLuint*)src)[j];
+                     }
                   }
-               }
-               break;
-            case GL_FLOAT:
-               if (ctx->Unpack.SwapBytes) {
-                  GLint value;
-                  for (j=0;j<elems_per_row;j++) {
-                     value = ((GLuint*)src)[j];
-                     value = ((value & 0xff000000) >> 24)
-                           | ((value & 0x00ff0000) >> 8)
-                           | ((value & 0x0000ff00) << 8)
-                           | ((value & 0x000000ff) << 24);
-                     *dst++ = *((GLfloat*) &value);
+                  break;
+               case GL_INT:
+                  if (ctx->Unpack.SwapBytes) {
+                     GLint value;
+                     for (j=0;j<elems_per_row;j++) {
+                        value = ((GLint*)src)[j];
+                        value = ((value & 0xff000000) >> 24)
+                              | ((value & 0x00ff0000) >> 8)
+                              | ((value & 0x0000ff00) << 8)
+                              | ((value & 0x000000ff) << 24);
+                        *dst++ = (GLfloat) value;
+                     }
                   }
-               }
-               else {
-                  MEMCPY( dst, src, elems_per_row*sizeof(GLfloat) );
-                  dst += elems_per_row;
-               }
-               break;
-            default:
-               abort();
-         } /*switch*/
+                  else {
+                     for (j=0;j<elems_per_row;j++) {
+                        *dst++ = (GLfloat) ((GLint*)src)[j];
+                     }
+                  }
+                  break;
+               case GL_FLOAT:
+                  if (ctx->Unpack.SwapBytes) {
+                     GLint value;
+                     for (j=0;j<elems_per_row;j++) {
+                        value = ((GLuint*)src)[j];
+                        value = ((value & 0xff000000) >> 24)
+                              | ((value & 0x00ff0000) >> 8)
+                              | ((value & 0x0000ff00) << 8)
+                              | ((value & 0x000000ff) << 24);
+                        *dst++ = *((GLfloat*) &value);
+                     }
+                  }
+                  else {
+                     MEMCPY( dst, src, elems_per_row*sizeof(GLfloat) );
+                     dst += elems_per_row;
+                  }
+                  break;
+               default:
+                  abort();
+            } /*switch*/
+         } /* for height */
       } /*for*/
 
       image = (struct gl_image *) malloc( sizeof(struct gl_image) );
       if (image) {
          image->Width = width;
          image->Height = height;
+         image->Depth = depth;  
          image->Components = components;
          image->Format = srcFormat;
          image->Type = GL_FLOAT;
