@@ -1,9 +1,9 @@
-/* $Id: teximage.c,v 1.7 1996/12/02 18:59:54 brianp Exp $ */
+/* $Id: teximage.c,v 1.14 1997/03/04 19:18:29 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  2.0
- * Copyright (C) 1995-1996  Brian Paul
+ * Version:  2.2
+ * Copyright (C) 1995-1997  Brian Paul
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +23,27 @@
 
 /*
  * $Log: teximage.c,v $
+ * Revision 1.14  1997/03/04 19:18:29  brianp
+ * added texture image Width2, Height2, and Depth2 fields
+ *
+ * Revision 1.13  1997/02/27 19:58:08  brianp
+ * call gl_problem() instead of gl_warning()
+ *
+ * Revision 1.12  1997/02/09 18:53:05  brianp
+ * added GL_EXT_texture3D support
+ *
+ * Revision 1.11  1997/01/16 03:35:10  brianp
+ * added calls to device driver TexImage() function
+ *
+ * Revision 1.10  1997/01/09 21:26:46  brianp
+ * gl_TexImage[12]D() didn't free the incoming image- a memory leak
+ *
+ * Revision 1.9  1997/01/09 19:55:52  brianp
+ * fixed a few error messages
+ *
+ * Revision 1.8  1997/01/09 19:49:18  brianp
+ * better error checking
+ *
  * Revision 1.7  1996/12/02 18:59:54  brianp
  * added code to handle GL_COLOR_INDEX textures, per Randy Frank
  *
@@ -77,7 +98,6 @@
  */
 static int logbase2( int n )
 {
-   GLint m = n;
    GLint i = 1;
    GLint log2 = 0;
 
@@ -85,11 +105,11 @@ static int logbase2( int n )
       return -1;
    }
 
-   while ( m > i ) {
+   while ( n > i ) {
       i *= 2;
       log2++;
    }
-   if (m != n) {
+   if (i != n) {
       return -1;
    }
    else {
@@ -250,6 +270,7 @@ void gl_free_texture_image( struct gl_texture_image *teximage )
  * Input:  image - the incoming gl_image
  *         internalFormat - desired format of resultant texture
  *         border - texture border width (0 or 1)
+ * Return:  pointer to a gl_texture_image or NULL if an error occurs.
  */
 struct gl_texture_image *gl_image_to_texture( GLcontext *ctx,
                                               const struct gl_image *image,
@@ -261,20 +282,39 @@ struct gl_texture_image *gl_image_to_texture( GLcontext *ctx,
    GLint numPixels, pixel;
    GLboolean scaleOrBias;
 
+   assert(image->Width>0);
+   assert(image->Height>0);
+   assert(image->Depth>0);
+
    internalFormat = decode_internal_format(internalFormat);
    components = components_in_intformat(internalFormat);
-   numPixels = image->Width * image->Height;
+   numPixels = image->Width * image->Height * image->Depth;
 
    texImage = gl_alloc_texture_image();
    texImage->Format = internalFormat;
    texImage->Border = border;
    texImage->Width = image->Width;
    texImage->Height = image->Height;
-   texImage->WidthLog2 = logbase2(image->Width);
-   texImage->HeightLog2 = logbase2(image->Height);
+   texImage->Depth = image->Depth;
+   texImage->WidthLog2 = logbase2(image->Width - 2*border);
+   if (image->Height==1)  /* 1-D texture */
+      texImage->HeightLog2 = 0;
+   else
+      texImage->HeightLog2 = logbase2(image->Height - 2*border);
+   if (image->Depth==1)   /* 2-D texture */
+      texImage->DepthLog2 = 0;
+   else
+      texImage->DepthLog2 = logbase2(image->Depth - 2*border);
+   texImage->Width2 = 1 << texImage->WidthLog2;
+   texImage->Height2 = 1 << texImage->HeightLog2;
+   texImage->Depth2 = 1 << texImage->DepthLog2;
    texImage->MaxLog2 = MAX2( texImage->WidthLog2, texImage->HeightLog2 );
    texImage->Data = (GLubyte *) malloc( numPixels * components );
 
+   assert(texImage->WidthLog2>=0);
+   assert(texImage->HeightLog2>=0);
+   assert(texImage->DepthLog2>=0);
+
    if (!texImage->Data) {
       /* out of memory */
       gl_free_texture_image( texImage );
@@ -441,7 +481,7 @@ struct gl_texture_image *gl_image_to_texture( GLcontext *ctx,
                /* Apply RGBA scale and bias */
                GLfloat r = red   * (1.0F/255.0F);
                GLfloat g = green * (1.0F/255.0F);
-               GLfloat b = blue * (1.0F/255.0F);
+               GLfloat b = blue  * (1.0F/255.0F);
                GLfloat a = alpha * (1.0F/255.0F);
                if (scaleOrBias) {
                   /* r,g,b,a now in [0,1] */
@@ -654,6 +694,7 @@ struct gl_texture_image *gl_image_to_texture( GLcontext *ctx,
 
 /*
  * Test glTexImagee1D() parameters for errors.
+ * Return:  GL_TRUE = an error was detected, GL_FALSE = no errors
  */
 static GLboolean texture_1d_error_check( GLcontext *ctx, GLenum target,
                                          GLint level, GLenum internalFormat,
@@ -674,21 +715,21 @@ static GLboolean texture_1d_error_check( GLcontext *ctx, GLenum target,
       gl_error( ctx, GL_INVALID_VALUE, "glTexImage1D(internalFormat)" );
       return GL_TRUE;
    }
-   if (width<2*border || width>2+MAX_TEXTURE_SIZE) {
-      if (target!=GL_PROXY_TEXTURE_1D) {
-         gl_error( ctx, GL_INVALID_VALUE, "glTexImage1D(width)" );
-      }
-      return GL_TRUE;
-   }
    if (border!=0 && border!=1) {
       if (target!=GL_PROXY_TEXTURE_1D) {
          gl_error( ctx, GL_INVALID_VALUE, "glTexImage1D(border)" );
       }
       return GL_TRUE;
    }
+   if (width<2*border || width>2+MAX_TEXTURE_SIZE) {
+      if (target!=GL_PROXY_TEXTURE_1D) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage1D(width)" );
+      }
+      return GL_TRUE;
+   }
    if (logbase2( width-2*border )<0) {
       gl_error( ctx, GL_INVALID_VALUE,
-               "glTexImage1D(width != 2^k + 2*border))");
+               "glTexImage1D(width != 2^k + 2*border)");
       return GL_TRUE;
    }
    switch (format) {
@@ -725,6 +766,7 @@ static GLboolean texture_1d_error_check( GLcontext *ctx, GLenum target,
 
 /*
  * Test glTexImagee2D() parameters for errors.
+ * Return:  GL_TRUE = an error was detected, GL_FALSE = no errors
  */
 static GLboolean texture_2d_error_check( GLcontext *ctx, GLenum target,
                                          GLint level, GLenum internalFormat,
@@ -746,6 +788,12 @@ static GLboolean texture_2d_error_check( GLcontext *ctx, GLenum target,
       gl_error( ctx, GL_INVALID_VALUE, "glTexImage2D(internalFormat)" );
       return GL_TRUE;
    }
+   if (border!=0 && border!=1) {
+      if (target!=GL_PROXY_TEXTURE_2D) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage2D(border)" );
+      }
+      return GL_TRUE;
+   }
    if (width<2*border || width>2+MAX_TEXTURE_SIZE) {
       if (target!=GL_PROXY_TEXTURE_2D) {
          gl_error( ctx, GL_INVALID_VALUE, "glTexImage2D(width)" );
@@ -758,20 +806,14 @@ static GLboolean texture_2d_error_check( GLcontext *ctx, GLenum target,
       }
       return GL_TRUE;
    }
-   if (border!=0 && border!=1) {
-      if (target!=GL_PROXY_TEXTURE_2D) {
-         gl_error( ctx, GL_INVALID_VALUE, "glTexImage2D(border)" );
-      }
-      return GL_TRUE;
-   }
    if (logbase2( width-2*border )<0) {
       gl_error( ctx,GL_INVALID_VALUE,
-               "glTexImage2D(width != 2^k + 2*border))");
+               "glTexImage2D(width != 2^k + 2*border)");
       return GL_TRUE;
    }
    if (logbase2( height-2*border )<0) {
       gl_error( ctx,GL_INVALID_VALUE,
-               "glTexImage2D(height != 2^k + 2*border))");
+               "glTexImage2D(height != 2^k + 2*border)");
       return GL_TRUE;
    }
    switch (format) {
@@ -808,6 +850,103 @@ static GLboolean texture_2d_error_check( GLcontext *ctx, GLenum target,
 }
 
 
+/*
+ * Test glTexImage3DEXT() parameters for errors.
+ * Return:  GL_TRUE = an error was detected, GL_FALSE = no errors
+ */
+static GLboolean texture_3d_error_check( GLcontext *ctx, GLenum target,
+                                         GLint level, GLenum internalFormat,
+                                         GLenum format, GLenum type,
+                                         GLint width, GLint height,
+                                         GLint depth, GLint border )
+{
+   GLint iformat;
+   if (target!=GL_TEXTURE_3D_EXT && target!=GL_PROXY_TEXTURE_3D_EXT) {
+      gl_error( ctx, GL_INVALID_ENUM, "glTexImage3DEXT(target)" );
+      return GL_TRUE;
+   }
+   if (level<0 || level>=MAX_TEXTURE_LEVELS) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(level)" );
+      return GL_TRUE;
+   }
+   iformat = decode_internal_format( internalFormat );
+   if (iformat<0) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(internalFormat)" );
+      return GL_TRUE;
+   }
+   if (border!=0 && border!=1) {
+      if (target!=GL_PROXY_TEXTURE_3D_EXT) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(border)" );
+      }
+      return GL_TRUE;
+   }
+   if (width<2*border || width>2+MAX_TEXTURE_SIZE) {
+      if (target!=GL_PROXY_TEXTURE_3D_EXT) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(width)" );
+      }
+      return GL_TRUE;
+   }
+   if (height<2*border || height>2+MAX_TEXTURE_SIZE) {
+      if (target!=GL_PROXY_TEXTURE_3D_EXT) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(height)" );
+      }
+      return GL_TRUE;
+   }
+   if (depth<2*border || depth>2+MAX_TEXTURE_SIZE) {
+      if (target!=GL_PROXY_TEXTURE_3D_EXT) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexImage3DEXT(depth)" );
+      }
+      return GL_TRUE;
+   }
+   if (logbase2( width-2*border )<0) {
+      gl_error( ctx,GL_INVALID_VALUE,
+               "glTexImage3DEXT(width != 2^k + 2*border))");
+      return GL_TRUE;
+   }
+   if (logbase2( height-2*border )<0) {
+      gl_error( ctx,GL_INVALID_VALUE,
+               "glTexImage3DEXT(height != 2^k + 2*border))");
+      return GL_TRUE;
+   }
+   if (logbase2( depth-2*border )<0) {
+      gl_error( ctx,GL_INVALID_VALUE,
+               "glTexImage3DEXT(depth  != 2^k + 2*border))");
+      return GL_TRUE;
+   }
+   switch (format) {
+      case GL_COLOR_INDEX:
+      case GL_RED:
+      case GL_GREEN:
+      case GL_BLUE:
+      case GL_ALPHA:
+      case GL_RGB:
+      case GL_RGBA:
+      case GL_LUMINANCE:
+      case GL_LUMINANCE_ALPHA:
+         /* OK */
+         break;
+      default:
+         gl_error( ctx, GL_INVALID_ENUM, "glTexImage3DEXT(format)" );
+         return GL_TRUE;
+   }
+   switch (type) {
+      case GL_UNSIGNED_BYTE:
+      case GL_BYTE:
+      case GL_UNSIGNED_SHORT:
+      case GL_SHORT:
+      case GL_UNSIGNED_INT:
+      case GL_INT:
+      case GL_FLOAT:
+         /* OK */
+         break;
+      default:
+         gl_error( ctx, GL_INVALID_ENUM, "glTexImage3DEXT(type)" );
+         return GL_TRUE;
+   }
+   return GL_FALSE;
+}
+
+
 
 /*
  * Called from the API.  Note that width includes the border.
@@ -823,17 +962,33 @@ void gl_TexImage1D( GLcontext *ctx,
    }
 
    if (image) {
-      /* if image is not NULL then it must be valid */
+      /* if image is not NULL then it's probably valid... */
       if (target==GL_TEXTURE_1D) {
          struct gl_texture_image *teximage;
+         if (texture_1d_error_check( ctx, target, level, internalformat,
+                                     format, type, width, border )) {
+            /* error in texture image was detected */
+            return;
+         }
          /* free current texture image, if any */
          if (ctx->Texture.Current1D->Image[level]) {
             gl_free_texture_image( ctx->Texture.Current1D->Image[level] );
          }
          /* install new texture image */
          teximage = gl_image_to_texture( ctx, image, internalformat, border );
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
          ctx->Texture.Current1D->Image[level] = teximage;
          ctx->NewState |= NEW_TEXTURING;
+
+         /* Send texture image to device driver */
+         if (ctx->Driver.TexImage) {
+            (*ctx->Driver.TexImage)( ctx, GL_TEXTURE_1D,
+                                     ctx->Texture.Current1D->Name,
+                                     level, internalformat, teximage );
+         }
       }
       else if (target==GL_PROXY_TEXTURE_1D) {
          /* Proxy texture: check for errors and update proxy state */
@@ -850,6 +1005,10 @@ void gl_TexImage1D( GLcontext *ctx,
             ctx->Texture.Proxy1D->Image[level]->Width = width;
             ctx->Texture.Proxy1D->Image[level]->Height = 1;
          }
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
       }
       else {
          gl_error( ctx, GL_INVALID_ENUM, "glTexImage1D(target)" );
@@ -884,17 +1043,33 @@ void gl_TexImage2D( GLcontext *ctx,
    }
 
    if (image) {
-      /* if image is not NULL then it must be valid */
+      /* if image is not NULL then it's probably valid... */
       if (target==GL_TEXTURE_2D) {
          struct gl_texture_image *teximage;
+         if (texture_2d_error_check( ctx, target, level, internalformat,
+                                     format, type, width, height, border )) {
+            /* error in texture image was detected */
+            return;
+         }
          /* free current texture image, if any */
          if (ctx->Texture.Current2D->Image[level]) {
             gl_free_texture_image( ctx->Texture.Current2D->Image[level] );
          }
          /* install new texture image */
          teximage = gl_image_to_texture( ctx, image, internalformat, border );
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
          ctx->Texture.Current2D->Image[level] = teximage;
          ctx->NewState |= NEW_TEXTURING;
+
+         /* Send texture image to device driver */
+         if (ctx->Driver.TexImage) {
+            (*ctx->Driver.TexImage)( ctx, GL_TEXTURE_2D,
+                                     ctx->Texture.Current2D->Name,
+                                     level, internalformat, teximage );
+         }
       }
       else if (target==GL_PROXY_TEXTURE_2D) {
          /* Proxy texture: check for errors and update proxy state */
@@ -911,6 +1086,10 @@ void gl_TexImage2D( GLcontext *ctx,
             ctx->Texture.Proxy2D->Image[level]->Width = width;
             ctx->Texture.Proxy2D->Image[level]->Height = height;
          }
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
       }
       else {
          gl_error( ctx, GL_INVALID_ENUM, "glTexImage2D(target)" );
@@ -929,6 +1108,86 @@ void gl_TexImage2D( GLcontext *ctx,
 
 
 
+/*
+ * Called by the API or display list executor.
+ * Note that width and height include the border.
+ */
+void gl_TexImage3DEXT( GLcontext *ctx,
+                       GLenum target, GLint level, GLint internalformat,
+                       GLsizei width, GLsizei height, GLsizei depth,
+                       GLint border, GLenum format, GLenum type,
+                       struct gl_image *image )
+{
+   if (INSIDE_BEGIN_END(ctx)) {
+      gl_error( ctx, GL_INVALID_OPERATION, "glTexImage3DEXT" );
+      return;
+   }
+
+   if (image) {
+      /* if image is not NULL then it must be valid */
+      if (target==GL_TEXTURE_3D_EXT) {
+         struct gl_texture_image *teximage;
+         if (texture_3d_error_check( ctx, target, level, internalformat,
+                                     format, type, width, height, depth,
+                                     border )) {
+            /* error in texture image was detected */
+            return;
+         }
+         /* free current texture image, if any */
+         if (ctx->Texture.Current3D->Image[level]) {
+            gl_free_texture_image( ctx->Texture.Current3D->Image[level] );
+         }
+         /* install new texture image */
+         teximage = gl_image_to_texture( ctx, image, internalformat, border );
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
+         ctx->Texture.Current3D->Image[level] = teximage;
+         ctx->NewState |= NEW_TEXTURING;
+
+         /* Send texture image to device driver */
+         /* TODO: device driver 3-D texture function */
+      }
+      else if (target==GL_PROXY_TEXTURE_3D_EXT) {
+         /* Proxy texture: check for errors and update proxy state */
+         if (texture_3d_error_check( ctx, target, level, internalformat,
+                                     format, type, width, height, depth,
+                                     border )) {
+            if (level>=0 && level<MAX_TEXTURE_LEVELS) {
+               MEMSET( ctx->Texture.Proxy3D->Image[level], 0,
+                      sizeof(struct gl_texture_image) );
+            }
+         }
+         else {
+            ctx->Texture.Proxy3D->Image[level]->Format = internalformat;
+            ctx->Texture.Proxy3D->Image[level]->Border = border;
+            ctx->Texture.Proxy3D->Image[level]->Width = width;
+            ctx->Texture.Proxy3D->Image[level]->Height = height;
+            ctx->Texture.Proxy3D->Image[level]->Depth  = depth;
+         }
+         if (image->RefCount==0) {
+            /* if RefCount==1 then image must be in a display list */
+            gl_free_image(image);
+         }
+      }
+      else {
+         gl_error( ctx, GL_INVALID_ENUM, "glTexImage3DEXT(target)" );
+         return;
+      }
+   }
+   else {
+      /*
+       * An error must have occured during texture unpacking.
+       * Take care of reporting the error now.
+       */
+      (void) texture_3d_error_check( ctx, target, level, internalformat,
+                                     format, type, width, height, depth,
+                                      border );
+   }
+}
+
+
 void gl_GetTexImage( GLcontext *ctx, GLenum target, GLint level, GLenum format,
                      GLenum type, GLvoid *pixels )
 {
@@ -984,7 +1243,44 @@ gl_unpack_texsubimage( GLcontext *ctx, GLint width, GLint height,
 }
 
 
+/*
+ * Unpack the image data given to glTexSubImage3D.
+ */
+struct gl_image *
+gl_unpack_texsubimage3D( GLcontext *ctx, GLint width, GLint height,
+                         GLint depth, GLenum format, GLenum type,
+                         const GLvoid *pixels )
+{
+   GLint components;
+   GLenum desttype;
 
+   if (width<0 || height<0 || depth<0 || !pixels) {
+      return NULL;
+   }
+
+   if (type==GL_BITMAP && format!=GL_COLOR_INDEX) {
+      return NULL;
+   }
+
+   components = components_in_intformat( format );
+   if (components<0 || format==GL_STENCIL_INDEX || format==GL_DEPTH_COMPONENT) {
+      return NULL;
+   }
+
+   if (gl_sizeof_type(type)<=0) {
+      return NULL;
+   }
+
+   if (type==GL_UNSIGNED_BYTE) {
+      desttype = GL_UNSIGNED_BYTE;
+   }
+   else {
+      desttype = GL_FLOAT;
+   }
+  
+   return gl_unpack_image3D( ctx, width, height, depth, format, type,
+                           desttype, pixels, GL_FALSE );
+}
 
 
 
@@ -1026,10 +1322,10 @@ void gl_TexSubImage1D( GLcontext *ctx,
 
       /* TODO: this is temporary.  If Type==GL_FLOAT or scale&bias needed */
       /* then do more work. */
-	  if (image->Type==GL_FLOAT) {
-		  gl_warning( ctx, "unimplemented texture type in glTexSubImage1D" );
-		  return;
-	  }
+      if (image->Type==GL_FLOAT) {
+         gl_problem( ctx, "unimplemented texture type in glTexSubImage1D" );
+         return;
+      }
 
       texcomponents = components_in_intformat(teximage->Format);
 
@@ -1040,7 +1336,7 @@ void gl_TexSubImage1D( GLcontext *ctx,
       }
       else {
          /* TODO: this is a hack */
-         gl_warning( ctx, "component mismatch in glTexSubImage1D" );
+         gl_problem( ctx, "component mismatch in glTexSubImage1D" );
          for (i=0;i<width;i++) {
             for (k=0;k<texcomponents;k++) {
                dst[k] = src[k];
@@ -1128,10 +1424,10 @@ void gl_TexSubImage2D( GLcontext *ctx,
 
       /* TODO: this is temporary.  If Type==GL_FLOAT or scale&bias needed */
       /* then do more work. */
-	  if (image->Type==GL_FLOAT) {
-		  gl_warning( ctx, "unimplemented texture type in glTexSubImage2D" );
-		  return;
-	  }
+      if (image->Type==GL_FLOAT) {
+         gl_problem( ctx, "unimplemented texture type in glTexSubImage2D" );
+         return;
+      }
 
       texcomponents = components_in_intformat(teximage->Format);
 
@@ -1147,7 +1443,7 @@ void gl_TexSubImage2D( GLcontext *ctx,
       }
       else {
          /* TODO: this is a hack */
-         gl_warning( ctx, "component mismatch in glTexSubImage2D" );
+         gl_problem( ctx, "component mismatch in glTexSubImage2D" );
 
          for (j=0;j<height;j++) {
             dst = teximage->Data 
@@ -1197,6 +1493,148 @@ void gl_TexSubImage2D( GLcontext *ctx,
 
 
 
+void gl_TexSubImage3DEXT( GLcontext *ctx,
+                          GLenum target, GLint level,
+                          GLint xoffset, GLint yoffset, GLint zoffset,
+                          GLsizei width, GLsizei height, GLsizei depth,
+                          GLenum format, GLenum type,
+                          struct gl_image *image )
+{
+   struct gl_texture_image *teximage;
+
+   if (target!=GL_TEXTURE_3D_EXT) {
+      gl_error( ctx, GL_INVALID_ENUM, "glTexSubImage3DEXT(target)" );
+      return;
+   }
+   if (level<0 || level>=MAX_TEXTURE_LEVELS) {
+      gl_error( ctx, GL_INVALID_ENUM, "glTexSubImage3DEXT(level)" );
+      return;
+   }
+
+   teximage = ctx->Texture.Current3D->Image[level];
+   if (!teximage) {
+      gl_error( ctx, GL_INVALID_OPERATION, "glTexSubImage3DEXT" );
+      return;
+   }
+
+   if (xoffset < -teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(xoffset)" );
+      return;
+   }
+   if (yoffset < -teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(yoffset)" );
+      return;
+   }
+   if (zoffset < -teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(zoffset)" );
+      return;
+   }
+   if (xoffset + width > teximage->Width+teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(xoffset+width)" );
+      return;
+   }
+   if (yoffset + height > teximage->Height+teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(yoffset+height)" );
+      return;
+   }
+   if (zoffset + depth  > teximage->Depth+teximage->Border) {
+      gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(zoffset+depth)" );
+      return;
+   }
+
+   if (image) {
+      /* unpacking must have been error-free */
+      GLint texcomponents, i, j, k, rectarea;
+      GLubyte *dst, *src;
+
+      /* TODO: this is temporary.  If Type==GL_FLOAT or scale&bias needed */
+      /* then do more work. */
+      if (image->Type==GL_FLOAT) {
+         gl_problem( ctx, "unimplemented texture type in glTexSubImagDEXT" );
+         return;
+      }
+
+      texcomponents = components_in_intformat(teximage->Format);
+      rectarea=teximage->Width * teximage->Height;
+
+      if (texcomponents == image->Components) {
+         dst = teximage->Data
+               + (zoffset * rectarea +  yoffset * teximage->Width + xoffset)
+               * texcomponents;
+         src = image->Data;
+         for(k=0;k<depth; k++) {
+           for (j=0;j<height;j++) {
+              MEMCPY( dst, src, width * texcomponents );
+              dst += teximage->Width * texcomponents;
+              src += width * texcomponents;
+           }
+           dst += rectarea * texcomponents;
+           src += rectarea * texcomponents;
+         }
+      }
+      else {
+         /* TODO: this is a hack */
+         GLint l, trarea, tzoffset, tloffset;
+
+         gl_problem( ctx, "component mismatch in glTexSubImage3DEXT" );
+         trarea=width*depth * image->Components;
+         for (l=0; l<depth; l++) {
+            tzoffset=(zoffset+l) * rectarea;
+            tloffset=l * trarea;
+            for (j=0;j<height;j++) {
+               dst = teximage->Data
+                   + (tzoffset + (yoffset+j) * teximage->Width + xoffset)
+                   * texcomponents;
+               src = (GLubyte *) image->Data + tloffset + j * width * image->Components;
+               for (i=0;i<width;i++) {
+                  for (k=0;k<texcomponents;k++) {
+                     dst[k] = src[k];
+                  }
+                  dst += texcomponents;
+                  src += image->Components;
+               }
+            }
+         }
+      }
+   }
+   else {
+      /* if no image, an error must have occured, do more testing now */
+      GLint components, size;
+
+      if (width<0) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(width)" );
+         return;
+      }
+      if (height<0) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(height)" );
+         return;
+      }
+      if (depth<0) {
+         gl_error( ctx, GL_INVALID_VALUE, "glTexSubImage3DEXT(depth)" );
+         return;
+      }
+      if (type==GL_BITMAP && format!=GL_COLOR_INDEX) {
+         gl_error( ctx, GL_INVALID_ENUM, "glTexSubImage3DEXT(format)" );
+         return;
+      }
+      components = components_in_intformat( format );
+      if (components<0 || format==GL_STENCIL_INDEX
+          || format==GL_DEPTH_COMPONENT){
+         gl_error( ctx, GL_INVALID_ENUM, "glTexSubImage3DEXT(format)" );
+         return;
+      }
+      size = gl_sizeof_type( type );
+      if (size<=0) {
+         gl_error( ctx, GL_INVALID_ENUM, "glTexSubImage3DEXT(type)" );
+         return;
+      }
+      /* if we get here, probably ran out of memory during unpacking */
+      gl_error( ctx, GL_OUT_OF_MEMORY, "glTexSubImage3DEXT" );
+   }
+}
+
+
+
 /*
  * Read an RGBA image from the frame buffer.
  * Input:  ctx - the context
@@ -1420,11 +1858,13 @@ void gl_CopyTexImage2D( GLcontext *ctx,
 static void copy_tex_sub_image( GLcontext *ctx, struct gl_texture_image *dest,
                                 GLint width, GLint height,
                                 GLint srcx, GLint srcy,
-                                GLint dstx, GLint dsty )
+                                GLint dstx, GLint dsty, GLint zoffset )
 {
    GLint i, j;
-   GLint format, components;
+   GLint format, components, rectarea;
 
+   rectarea = width*height;
+   zoffset *= rectarea; 
    format = dest->Format;
    components = components_in_intformat( format );
 
@@ -1449,7 +1889,7 @@ static void copy_tex_sub_image( GLcontext *ctx, struct gl_texture_image *dest,
          }
       }
 
-      texptr = dest->Data + ((dsty+j) * width + dstx) * components;
+      texptr = dest->Data + ( zoffset + (dsty+j) * width + dstx) * components;
 
       switch (format) {
          case GL_ALPHA:
@@ -1532,7 +1972,7 @@ void gl_CopyTexSubImage1D( GLcontext *ctx,
          return;
       }
       if (teximage->Data) {
-         copy_tex_sub_image( ctx, teximage, width, 1, x, y, xoffset, 0 );
+         copy_tex_sub_image( ctx, teximage, width, 1, x, y, xoffset, 0, 0 );
       }
    }
    else {
@@ -1595,10 +2035,82 @@ void gl_CopyTexSubImage2D( GLcontext *ctx,
 
       if (teximage->Data) {
          copy_tex_sub_image( ctx, teximage, width, height,
-                             x, y, xoffset, yoffset );
+                             x, y, xoffset, yoffset, 0 );
       }
    }
    else {
       gl_error( ctx, GL_INVALID_OPERATION, "glCopyTexSubImage2D" );
    }
 }
+
+
+
+void gl_CopyTexSubImage3DEXT( GLcontext *ctx,
+                              GLenum target, GLint level,
+                              GLint xoffset, GLint yoffset, GLint zoffset,
+                              GLint x, GLint y, GLsizei width, GLsizei height )
+{
+   struct gl_texture_image *teximage;
+
+   if (INSIDE_BEGIN_END(ctx)) {
+      gl_error( ctx, GL_INVALID_OPERATION, "glCopyTexSubImage3DEXT" );
+      return;
+   }
+   if (target!=GL_TEXTURE_2D) {
+      gl_error( ctx, GL_INVALID_ENUM, "glCopyTexSubImage3DEXT(target)" );
+      return;
+   }
+   if (level<0 || level>=MAX_TEXTURE_LEVELS) {
+      gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(level)" );
+      return;
+   }
+   if (width<0) {
+      gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(width)" );
+      return;
+   }
+   if (height<0) {
+      gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(height)" );
+      return;
+   }
+
+   teximage = ctx->Texture.Current3D->Image[level];
+   if (teximage) {
+      if (xoffset < -teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(xoffset)" );
+         return;
+      }
+      if (yoffset < -teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(yoffset)" );
+         return;
+      }
+      if (zoffset < -teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE, "glCopyTexSubImage3DEXT(zoffset)" );
+         return;
+      }
+      /* NOTE: we're adding the border here, not subtracting! */
+      if (xoffset+width > teximage->Width+teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE,
+                   "glCopyTexSubImage3DEXT(xoffset+width)" );
+         return;
+      }
+      if (yoffset+height > teximage->Height+teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE,
+                   "glCopyTexSubImage3DEXT(yoffset+height)" );
+         return;
+      }
+      if (zoffset > teximage->Depth+teximage->Border) {
+         gl_error( ctx, GL_INVALID_VALUE,
+                   "glCopyTexSubImage3DEXT(zoffset+depth)" );
+         return;
+      }
+
+      if (teximage->Data) {
+         copy_tex_sub_image( ctx, teximage, width, height, 
+                             x, y, xoffset, yoffset, zoffset);
+      }
+   }
+   else {
+      gl_error( ctx, GL_INVALID_OPERATION, "glCopyTexSubImage3DEXT" );
+   }
+}
+
