@@ -1,9 +1,9 @@
-/* $Id: triangle.c,v 1.5 1996/10/01 03:31:17 brianp Exp $ */
+/* $Id: triangle.c,v 1.17 1997/03/13 03:05:31 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  2.0
- * Copyright (C) 1995-1996  Brian Paul
+ * Version:  2.2
+ * Copyright (C) 1995-1997  Brian Paul
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +23,43 @@
 
 /*
  * $Log: triangle.c,v $
+ * Revision 1.17  1997/03/13 03:05:31  brianp
+ * removed unused shift variable in feedback_triangle()
+ *
+ * Revision 1.16  1997/03/08 02:04:27  brianp
+ * better implementation of feedback function
+ *
+ * Revision 1.15  1997/03/04 18:54:13  brianp
+ * renamed mipmap_textured_triangle() to lambda_textured_triangle()
+ * better comments about lambda and mipmapping
+ *
+ * Revision 1.14  1997/02/20 23:47:35  brianp
+ * triangle feedback colors were wrong when using smooth shading
+ *
+ * Revision 1.13  1997/02/19 10:24:26  brianp
+ * use a GLdouble instead of a GLfloat for wwvvInv (perspective correction)
+ *
+ * Revision 1.12  1997/02/09 19:53:43  brianp
+ * now use TEXTURE_xD enable constants
+ *
+ * Revision 1.11  1997/02/09 18:51:02  brianp
+ * added GL_EXT_texture3D support
+ *
+ * Revision 1.10  1997/01/16 03:36:43  brianp
+ * added #include "texture.h"
+ *
+ * Revision 1.9  1997/01/09 19:50:49  brianp
+ * now call gl_texturing_enabled()
+ *
+ * Revision 1.8  1996/12/20 20:23:30  brianp
+ * the test for using general_textured_triangle() was wrong
+ *
+ * Revision 1.7  1996/12/12 22:37:30  brianp
+ * projective textures didn't work right
+ *
+ * Revision 1.6  1996/11/08 02:21:21  brianp
+ * added null drawing function for GL_NO_RASTER
+ *
  * Revision 1.5  1996/10/01 03:31:17  brianp
  * use new FixedToDepth() macro
  *
@@ -53,6 +90,7 @@
 #include "feedback.h"
 #include "macros.h"
 #include "span.h"
+#include "texture.h"
 #include "triangle.h"
 #include "types.h"
 #include "vb.h"
@@ -66,15 +104,30 @@ static void feedback_triangle( GLcontext *ctx,
                                GLuint v0, GLuint v1, GLuint v2, GLuint pv )
 {
    struct vertex_buffer *VB = ctx->VB;
-   GLfloat x, y, z, w;
-   GLfloat tc[4];
    GLfloat color[4];
-   GLuint i, v;
+   GLuint i;
+   GLfloat invRedScale   = ctx->Visual->InvRedScale;
+   GLfloat invGreenScale = ctx->Visual->InvGreenScale;
+   GLfloat invBlueScale  = ctx->Visual->InvBlueScale;
+   GLfloat invAlphaScale = ctx->Visual->InvAlphaScale;
 
    FEEDBACK_TOKEN( ctx, (GLfloat) GL_POLYGON_TOKEN );
    FEEDBACK_TOKEN( ctx, (GLfloat) 3 );        /* three vertices */
 
+   if (ctx->Light.ShadeModel==GL_FLAT) {
+      /* flat shading - same color for each vertex */
+      color[0] = (GLfloat) VB->Color[pv][0] * invRedScale;
+      color[1] = (GLfloat) VB->Color[pv][1] * invGreenScale;
+      color[2] = (GLfloat) VB->Color[pv][2] * invBlueScale;
+      color[3] = (GLfloat) VB->Color[pv][3] * invAlphaScale;
+   }
+
    for (i=0;i<3;i++) {
+      GLfloat x, y, z, w;
+      GLfloat tc[4];
+      GLuint v;
+      GLfloat invq;
+
       if (i==0)       v = v0;
       else if (i==1)  v = v1;
       else            v = v2;
@@ -84,16 +137,19 @@ static void feedback_triangle( GLcontext *ctx,
       z = VB->Win[v][2] / DEPTH_SCALE;
       w = VB->Clip[v][3];
 
-      /* convert color from integer back to a float in [0,1] */
-      color[0] = (GLfloat) VB->Color[v][0] * ctx->Visual->InvRedScale;
-      color[1] = (GLfloat) VB->Color[v][1] * ctx->Visual->InvGreenScale;
-      color[2] = (GLfloat) VB->Color[v][2] * ctx->Visual->InvBlueScale;
-      color[3] = (GLfloat) VB->Color[v][3] * ctx->Visual->InvAlphaScale;
+      if (ctx->Light.ShadeModel==GL_SMOOTH) {
+         /* smooth shading - different color for each vertex */
+         color[0] = FixedToFloat(VB->Color[v][0]) * invRedScale;
+         color[1] = FixedToFloat(VB->Color[v][1]) * invGreenScale;
+         color[2] = FixedToFloat(VB->Color[v][2]) * invBlueScale;
+         color[3] = FixedToFloat(VB->Color[v][3]) * invAlphaScale;
+      }
 
-      tc[0] = VB->TexCoord[v][0];
-      tc[1] = VB->TexCoord[v][1];
-      tc[2] = 0.0F;          /* TODO: R, Q components */
-      tc[3] = 1.0F;
+      invq = 1.0F / VB->TexCoord[v][3];
+      tc[0] = VB->TexCoord[v][0] * invq;
+      tc[1] = VB->TexCoord[v][1] * invq;
+      tc[2] = VB->TexCoord[v][2] * invq;
+      tc[3] = VB->TexCoord[v][3];
 
       gl_feedback_vertex( ctx, x, y, z, w, color, (GLfloat) VB->Index[v], tc );
    }
@@ -366,7 +422,9 @@ static void simple_z_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
 
 /*
  * Render a smooth-shaded, textured, RGBA triangle.
- * Interpolate S,T with perspective correction, w/out mipmapping.
+ * Interpolate S,T,U with perspective correction, w/out mipmapping.
+ * Note: we use texture coordinates S,T,U,V instead of S,T,R,Q because
+ * R is already used for red.
  */
 static void general_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
                                        GLuint v2, GLuint pv )
@@ -375,8 +433,7 @@ static void general_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
 #define INTERP_RGB 1
 #define INTERP_ALPHA 1
 #define INTERP_STW 1
-#define S_SCALE 100.0F
-#define T_SCALE 100.0F
+#define INTERP_UV 1
 #define SETUP_CODE						\
    GLboolean flat_shade = (ctx->Light.ShadeModel==GL_FLAT);	\
    GLint r, g, b, a;						\
@@ -392,32 +449,38 @@ static void general_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
 	   GLdepth zspan[MAX_WIDTH];				\
 	   GLubyte red[MAX_WIDTH], green[MAX_WIDTH];		\
 	   GLubyte blue[MAX_WIDTH], alpha[MAX_WIDTH];		\
-           GLfloat s[MAX_WIDTH], t[MAX_WIDTH];			\
+           GLfloat s[MAX_WIDTH], t[MAX_WIDTH], u[MAX_WIDTH];	\
 	   if (n>0) {						\
               if (flat_shade) {					\
                  for (i=0;i<n;i++) {				\
+		    GLdouble wwvvInv = 1.0 / (ww*vv);		\
 		    zspan[i] = FixedToDepth(ffz);		\
 		    red[i]   = r;				\
 		    green[i] = g;				\
 		    blue[i]  = b;				\
 		    alpha[i] = a;				\
-		    s[i] = ss/ww;				\
-		    t[i] = tt/ww;				\
+		    s[i] = ss*wwvvInv;				\
+		    t[i] = tt*wwvvInv;				\
+		    u[i] = uu*wwvvInv;				\
 		    ffz += fdzdx;				\
 		    ss += dsdx;					\
 		    tt += dtdx;					\
+		    uu += dudx;					\
+		    vv += dvdx;					\
 		    ww += dwdx;					\
 		 }						\
               }							\
               else {						\
                  for (i=0;i<n;i++) {				\
+		    GLdouble wwvvInv = 1.0 / (ww*vv);		\
 		    zspan[i] = FixedToDepth(ffz);		\
 		    red[i]   = FixedToInt(ffr);			\
 		    green[i] = FixedToInt(ffg);			\
 		    blue[i]  = FixedToInt(ffb);			\
 		    alpha[i] = FixedToInt(ffa);			\
-		    s[i] = ss/ww;				\
-		    t[i] = tt/ww;				\
+		    s[i] = ss*wwvvInv;				\
+		    t[i] = tt*wwvvInv;				\
+		    u[i] = uu*wwvvInv;				\
 		    ffz += fdzdx;				\
 		    ffr += fdrdx;				\
 		    ffg += fdgdx;				\
@@ -425,11 +488,13 @@ static void general_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
 		    ffa += fdadx;				\
 		    ss += dsdx;					\
 		    tt += dtdx;					\
+		    uu += dudx;					\
 		    ww += dwdx;					\
+		    vv += dvdx;					\
 		 }						\
               }							\
 	      gl_write_texture_span( ctx, n, LEFT, Y, zspan,	\
-                                     s, t, NULL, 		\
+                                     s, t, u, NULL, 		\
 	                             red, green, blue, alpha,	\
 				     GL_POLYGON );		\
 	   }							\
@@ -490,21 +555,36 @@ static GLfloat compute_lambda( GLfloat s, GLfloat t,
 
 /*
  * Render a smooth-shaded, textured, RGBA triangle.
- * Interpolate S,T with perspective correction and mipmapping.
+ * Interpolate S,T,U with perspective correction and compute lambda for
+ * each fragment.  Lambda is used to determine whether to use the
+ * minification or magnification filter.  If minification and using
+ * mipmaps, lambda is also used to select the texture level of detail.
  */
-static void mipmap_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
+static void lambda_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
                                       GLuint v2, GLuint pv )
 {
 #define INTERP_Z 1
 #define INTERP_RGB 1
 #define INTERP_ALPHA 1
 #define INTERP_STW 1
+#define INTERP_UV 1
 
 #define SETUP_CODE							\
-   GLfloat twidth = (GLfloat) ctx->Texture.Current2D->Image[0]->Width;	\
-   GLfloat theight = (GLfloat) ctx->Texture.Current2D->Image[0]->Height;\
    GLboolean flat_shade = (ctx->Light.ShadeModel==GL_FLAT);		\
    GLint r, g, b, a;							\
+   GLfloat twidth, theight;						\
+   if (ctx->Texture.Enabled & TEXTURE_3D) {				\
+      twidth = (GLfloat) ctx->Texture.Current3D->Image[0]->Width;	\
+      theight = (GLfloat) ctx->Texture.Current3D->Image[0]->Height;	\
+   }									\
+   else if (ctx->Texture.Enabled & TEXTURE_2D) {			\
+      twidth = (GLfloat) ctx->Texture.Current2D->Image[0]->Width;	\
+      theight = (GLfloat) ctx->Texture.Current2D->Image[0]->Height;	\
+   }									\
+   else {								\
+      twidth = (GLfloat) ctx->Texture.Current1D->Image[0]->Width;	\
+      theight = 1.0;							\
+   }									\
    if (flat_shade) {							\
       r = VB->Color[pv][0];						\
       g = VB->Color[pv][1];						\
@@ -512,62 +592,70 @@ static void mipmap_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
       a = VB->Color[pv][3];						\
    }
 
-#define INNER_LOOP( LEFT, RIGHT, Y )				\
-	{							\
-	   GLint i, n = RIGHT-LEFT;				\
-	   GLdepth zspan[MAX_WIDTH];				\
-	   GLubyte red[MAX_WIDTH], green[MAX_WIDTH];		\
-	   GLubyte blue[MAX_WIDTH], alpha[MAX_WIDTH];		\
-           GLfloat s[MAX_WIDTH], t[MAX_WIDTH];			\
-           GLfloat lambda[MAX_WIDTH];				\
-	   if (n>0) {						\
-	      if (flat_shade) {					\
-		 for (i=0;i<n;i++) {				\
-		    zspan[i] = FixedToDepth(ffz);		\
-		    red[i]   = r;				\
-		    green[i] = g;				\
-		    blue[i]  = b;				\
-		    alpha[i] = a;				\
-		    s[i] = ss/ww;				\
-		    t[i] = tt/ww;				\
-		    lambda[i] = compute_lambda( s[i], t[i],	\
-						dsdx, dsdy,	\
-						dtdx, dtdy, ww,	\
+#define INNER_LOOP( LEFT, RIGHT, Y )					\
+	{								\
+	   GLint i, n = RIGHT-LEFT;					\
+	   GLdepth zspan[MAX_WIDTH];					\
+	   GLubyte red[MAX_WIDTH], green[MAX_WIDTH];			\
+	   GLubyte blue[MAX_WIDTH], alpha[MAX_WIDTH];			\
+           GLfloat s[MAX_WIDTH], t[MAX_WIDTH], u[MAX_WIDTH];		\
+           GLfloat lambda[MAX_WIDTH];					\
+	   if (n>0) {							\
+	      if (flat_shade) {						\
+		 for (i=0;i<n;i++) {					\
+		    GLdouble wwvvInv = 1.0 / (ww*vv);			\
+		    zspan[i] = FixedToDepth(ffz);			\
+		    red[i]   = r;					\
+		    green[i] = g;					\
+		    blue[i]  = b;					\
+		    alpha[i] = a;					\
+		    s[i] = ss*wwvvInv;					\
+		    t[i] = tt*wwvvInv;					\
+		    u[i] = uu*wwvvInv;					\
+		    lambda[i] = compute_lambda( s[i], t[i],		\
+						dsdx, dsdy,		\
+						dtdx, dtdy, ww,		\
 						twidth, theight );	\
-		    ffz += fdzdx;				\
-		    ss += dsdx;					\
-		    tt += dtdx;					\
-		    ww += dwdx;					\
-		 }						\
-              }							\
-              else {						\
-		 for (i=0;i<n;i++) {				\
-		    zspan[i] = FixedToDepth(ffz);		\
-		    red[i]   = FixedToInt(ffr);			\
-		    green[i] = FixedToInt(ffg);			\
-		    blue[i]  = FixedToInt(ffb);			\
-		    alpha[i] = FixedToInt(ffa);			\
-		    s[i] = ss/ww;				\
-		    t[i] = tt/ww;				\
-		    lambda[i] = compute_lambda( s[i], t[i],	\
-						dsdx, dsdy,	\
-						dtdx, dtdy, ww,	\
+		    ffz += fdzdx;					\
+		    ss += dsdx;						\
+		    tt += dtdx;						\
+		    uu += dudx;						\
+		    vv += dvdx;						\
+		    ww += dwdx;						\
+		 }							\
+              }								\
+              else {							\
+		 for (i=0;i<n;i++) {					\
+		    GLdouble wwvvInv = 1.0 / (ww*vv);			\
+		    zspan[i] = FixedToDepth(ffz);			\
+		    red[i]   = FixedToInt(ffr);				\
+		    green[i] = FixedToInt(ffg);				\
+		    blue[i]  = FixedToInt(ffb);				\
+		    alpha[i] = FixedToInt(ffa);				\
+		    s[i] = ss*wwvvInv;					\
+		    t[i] = tt*wwvvInv;					\
+		    u[i] = uu*wwvvInv;					\
+		    lambda[i] = compute_lambda( s[i], t[i],		\
+						dsdx, dsdy,		\
+						dtdx, dtdy, ww,		\
 						twidth, theight );	\
-		    ffz += fdzdx;				\
-		    ffr += fdrdx;				\
-		    ffg += fdgdx;				\
-		    ffb += fdbdx;				\
-		    ffa += fdadx;				\
-		    ss += dsdx;					\
-		    tt += dtdx;					\
-		    ww += dwdx;					\
-		 }						\
-              }							\
-	      gl_write_texture_span( ctx, n, LEFT, Y, zspan,	\
-                                     s, t, lambda, 		\
-	                             red, green, blue, alpha,	\
-				     GL_POLYGON );		\
-	   }							\
+		    ffz += fdzdx;					\
+		    ffr += fdrdx;					\
+		    ffg += fdgdx;					\
+		    ffb += fdbdx;					\
+		    ffa += fdadx;					\
+		    ss += dsdx;						\
+		    tt += dtdx;						\
+		    uu += dtdx;						\
+		    vv += dvdx;						\
+		    ww += dwdx;						\
+		 }							\
+              }								\
+	      gl_write_texture_span( ctx, n, LEFT, Y, zspan,		\
+                                     s, t, u, lambda,	 		\
+	                             red, green, blue, alpha,		\
+				     GL_POLYGON );			\
+	   }								\
 	}
 
 #include "tritemp.h"
@@ -575,6 +663,16 @@ static void mipmap_textured_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
 
 
 
+/*
+ * Null rasterizer for measuring transformation speed.
+ */
+static void null_triangle( GLcontext *ctx, GLuint v0, GLuint v1,
+                           GLuint v2, GLuint pv )
+{
+}
+
+
+
 /*
  * Determine which triangle rendering function to use given the current
  * rendering context.
@@ -584,21 +682,27 @@ void gl_set_triangle_function( GLcontext *ctx )
    GLboolean rgbmode = ctx->Visual->RGBAflag;
 
    if (ctx->RenderMode==GL_RENDER) {
+      if (ctx->NoRaster) {
+         ctx->TriangleFunc = null_triangle;
+         return;
+      }
       if (ctx->Driver.TriangleFunc) {
          /* Device driver will draw triangles. */
          ctx->TriangleFunc = ctx->Driver.TriangleFunc;
       }
-      else if (ctx->Texture.Enabled) {
-         if (   (ctx->Texture.Enabled & 2)
+      else if (gl_texturing_enabled(ctx)) {
+         if (   (ctx->Texture.Enabled==TEXTURE_2D)
              && ctx->Texture.Current2D->Complete
              && ctx->Texture.Current2D->MinFilter==GL_NEAREST
              && ctx->Texture.Current2D->MagFilter==GL_NEAREST
              && ctx->Texture.Current2D->WrapS==GL_REPEAT
              && ctx->Texture.Current2D->WrapT==GL_REPEAT
              && ctx->Texture.Current2D->Image[0]->Format==GL_RGB
+             && ctx->Texture.Current2D->Image[0]->Border==0
              && (ctx->Texture.EnvMode==GL_DECAL
                  || ctx->Texture.EnvMode==GL_REPLACE)
              && ctx->Hint.PerspectiveCorrection==GL_FASTEST
+             && ctx->IdentityTexMat
              && (ctx->RasterMask==DEPTH_BIT || ctx->RasterMask==0)
              && ctx->Depth.Func==GL_LESS
              && ctx->Depth.Mask==GL_TRUE
@@ -612,21 +716,28 @@ void gl_set_triangle_function( GLcontext *ctx )
             }
          }
          else {
-            GLboolean mipmap = GL_TRUE;
-            if (ctx->Texture.Enabled & 2) {
-               if (   ctx->Texture.Current2D->MinFilter==GL_NEAREST
-                      || ctx->Texture.Current2D->MinFilter==GL_LINEAR) {
-                  mipmap = GL_FALSE;
+            GLboolean needLambda = GL_TRUE;
+            /* if mag filter == min filter we're not mipmapping */
+            if (ctx->Texture.Enabled & TEXTURE_3D) {
+               if (ctx->Texture.Current3D->MinFilter==
+                   ctx->Texture.Current3D->MagFilter) {
+                  needLambda = GL_FALSE;
+               }
+	    }
+            else if (ctx->Texture.Enabled & TEXTURE_2D) {
+               if (ctx->Texture.Current2D->MinFilter==
+                   ctx->Texture.Current2D->MagFilter) {
+                  needLambda = GL_FALSE;
                }
             }
-            else {
-               if (   ctx->Texture.Current1D->MinFilter==GL_NEAREST 
-                      || ctx->Texture.Current1D->MinFilter==GL_LINEAR) {
-                  mipmap = GL_FALSE;
+            else if (ctx->Texture.Enabled & TEXTURE_1D) {
+               if (ctx->Texture.Current1D->MinFilter==
+                   ctx->Texture.Current1D->MagFilter) {
+                  needLambda = GL_FALSE;
                }
             }
-            ctx->TriangleFunc = mipmap ? mipmap_textured_triangle
-                                       : general_textured_triangle;
+            ctx->TriangleFunc = needLambda ? lambda_textured_triangle
+                                           : general_textured_triangle;
          }
       }
       else {
