@@ -1,9 +1,9 @@
-/* $Id: types.h,v 1.4 1996/09/25 03:22:53 brianp Exp $ */
+/* $Id: types.h,v 1.19 1997/03/13 03:05:54 brianp Exp $ */
 
 /*
  * Mesa 3-D graphics library
- * Version:  2.0
- * Copyright (C) 1995-1996  Brian Paul
+ * Version:  2.2
+ * Copyright (C) 1995-1997  Brian Paul
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +23,51 @@
 
 /*
  * $Log: types.h,v $
+ * Revision 1.19  1997/03/13 03:05:54  brianp
+ * changed AlphaRefInt to AlphaRefUbyte
+ *
+ * Revision 1.18  1997/03/04 19:17:38  brianp
+ * added a few members to gl_texture_image struct
+ *
+ * Revision 1.17  1997/02/10 19:49:29  brianp
+ * added glResizeBuffersMESA() code
+ *
+ * Revision 1.16  1997/02/09 19:53:43  brianp
+ * now use TEXTURE_xD enable constants
+ *
+ * Revision 1.15  1997/02/09 18:42:41  brianp
+ * added GL_EXT_texture3D support
+ *
+ * Revision 1.14  1997/01/28 22:14:36  brianp
+ * now there's separate state for CI and RGBA logic op enabled
+ *
+ * Revision 1.13  1997/01/21 03:09:41  brianp
+ * added definition:  union node;
+ *
+ * Revision 1.12  1997/01/09 21:26:23  brianp
+ * added RefCount to gl_image struct
+ *
+ * Revision 1.11  1996/12/18 20:01:39  brianp
+ * added material bitmask constants and ColorMaterialBitmask
+ *
+ * Revision 1.10  1996/12/11 20:16:30  brianp
+ * changed GLaccum types to signed
+ *
+ * Revision 1.9  1996/11/08 02:21:04  brianp
+ * added NoRaster flag to context
+ *
+ * Revision 1.8  1996/11/07 04:11:34  brianp
+ * now pass gl_image struct pointer to glTexImage[12]D()
+ *
+ * Revision 1.7  1996/11/06 04:21:58  brianp
+ * added Format to struct gl_image
+ *
+ * Revision 1.6  1996/10/11 03:41:12  brianp
+ * removed OffsetBias field
+ *
+ * Revision 1.5  1996/10/11 00:23:04  brianp
+ * changed Polygon/Line/PointZoffset to GLfloat
+ *
  * Revision 1.4  1996/09/25 03:22:53  brianp
  * added NO_DRAW_BIT for glDrawBuffer(GL_NONE)
  *
@@ -51,9 +96,9 @@
  * Accumulation buffer data type:
  */
 #if ACCUM_BITS==8
-   typedef GLubyte GLaccum;
+   typedef GLbyte GLaccum;
 #elif ACCUM_BITS==16
-   typedef GLushort GLaccum;
+   typedef GLshort GLaccum;
 #else
    illegal number of accumulation bits
 #endif
@@ -115,7 +160,9 @@ typedef void (*triangle_func)( GLcontext *ctx,
 struct gl_image {
 	GLint Width;
 	GLint Height;
+	GLint Depth;		/* for 3-D texturing */
 	GLint Components;	/* 1, 2, 3 or 4 */
+        GLenum Format;		/* GL_COLOR_INDEX, GL_RED, GL_RGB, etc */
 	GLenum Type;		/* GL_UNSIGNED_BYTE or GL_FLOAT or GL_BITMAP */
 	GLvoid *Data;
 	GLboolean Interleaved;	/* If TRUE and Format==GL_RGB, GL_RGBA or
@@ -124,6 +171,7 @@ struct gl_image {
 				 * RGBRGBRGBRGB..RGB.  This is only used
 				 * for glDrawPixels.
 				 */
+	GLint RefCount;
 };
 
 
@@ -134,13 +182,17 @@ struct gl_texture_image {
 				 * GL_INTENSITY, GL_RGB, or GL_RGBA
 				 */
 	GLuint Border;		/* 0 or 1 */
-	GLuint Width;		/* = 2^WidthLog2 */
-	GLuint Height;		/* = 2^HeightLog2 */
-	GLuint WidthLog2;	/* = log2(Width) */
-	GLuint HeightLog2;	/* = log2(Height) */
+	GLuint Width;		/* = 2^WidthLog2 + 2*Border */
+	GLuint Height;		/* = 2^HeightLog2 + 2*Border */
+	GLuint Depth;		/* = 2^DepthLog2 + 2*Border */
+	GLuint Width2;		/* = Width - 2*Border */
+	GLuint Height2;		/* = Height - 2*Border */
+	GLuint Depth2;		/* = Depth - 2*Border */
+	GLuint WidthLog2;	/* = log2(Width2) */
+	GLuint HeightLog2;	/* = log2(Height2) */
+	GLuint DepthLog2;	/* = log2(Depth2) */
 	GLuint MaxLog2;		/* = MAX(WidthLog2, HeightLog2) */
 	GLubyte *Data;		/* Image data as unsigned bytes */
-	GLboolean DeleteFlag;
 };
 
 
@@ -185,6 +237,11 @@ struct api_function_table {
                               GLint, GLint, GLsizei );
    void (*CopyTexSubImage2D)( GLcontext *, GLenum, GLint, GLint, GLint,
                               GLint, GLint, GLsizei, GLsizei );
+   void (*CopyTexSubImage3DEXT)(GLcontext *,
+                                GLenum , GLint ,
+                                GLint , GLint , GLint ,
+                                GLint , GLint ,
+                                GLsizei , GLsizei  );
    void (*CullFace)( GLcontext *, GLenum );
    void (*DeleteLists)( GLcontext *, GLuint, GLsizei );
    void (*DeleteTextures)( GLcontext *, GLsizei, const GLuint *);
@@ -328,14 +385,26 @@ struct api_function_table {
    void (*TexEnvfv)( GLcontext *, GLenum, GLenum, const GLfloat * );
    void (*TexGenfv)( GLcontext *, GLenum coord, GLenum, const GLfloat * );
    void (*TexImage1D)( GLcontext *, GLenum, GLint, GLint, GLsizei,
-                       GLint, GLenum, GLenum, struct gl_texture_image * );
+                       GLint, GLenum, GLenum, struct gl_image * );
    void (*TexImage2D)( GLcontext *, GLenum, GLint, GLint, GLsizei, GLsizei,
-			 GLint, GLenum, GLenum, struct gl_texture_image * );
+                       GLint, GLenum, GLenum, struct gl_image * );
    void (*TexSubImage1D)( GLcontext *, GLenum, GLint, GLint, GLsizei,
                           GLenum, GLenum, struct gl_image * );
    void (*TexSubImage2D)( GLcontext *, GLenum, GLint, GLint, GLint,
                           GLsizei, GLsizei, GLenum, GLenum,
-                          struct gl_image *image );
+                          struct gl_image * );
+   void (*TexImage3DEXT)(GLcontext *,
+                         GLenum , GLint , GLint ,
+                         GLsizei , GLsizei , GLsizei ,
+                         GLint ,
+                         GLenum , GLenum ,
+                         struct gl_image * );
+   void (*TexSubImage3DEXT)(GLcontext *,
+                            GLenum , GLint ,
+                            GLint , GLint , GLint,
+                            GLsizei , GLsizei , GLsizei ,
+                            GLenum , GLenum ,
+                            struct gl_image * );
    void (*TexParameterfv)( GLcontext *, GLenum, GLenum, const GLfloat * );
    /* Translated implemented by Translatef */
    void (*Translatef)( GLcontext *, GLfloat, GLfloat, GLfloat );
@@ -345,6 +414,9 @@ struct api_function_table {
 
    /* GL_MESA_window_pos extension */
    void (*WindowPos4fMESA)( GLcontext *, GLfloat, GLfloat, GLfloat, GLfloat );
+
+   /* GL_MESA_resize_buffers extension */
+   void (*ResizeBuffersMESA)( GLcontext * );
 };
 
 
@@ -352,6 +424,34 @@ struct api_function_table {
 #include "dd.h"
 
 
+/*
+ * Bit flags used for updating material values.
+ */
+#define FRONT_AMBIENT_BIT     0x1
+#define BACK_AMBIENT_BIT      0x2
+#define FRONT_DIFFUSE_BIT     0x4
+#define BACK_DIFFUSE_BIT      0x8
+#define FRONT_SPECULAR_BIT   0x10
+#define BACK_SPECULAR_BIT    0x20
+#define FRONT_EMISSION_BIT   0x40
+#define BACK_EMISSION_BIT    0x80
+#define FRONT_SHININESS_BIT 0x100
+#define BACK_SHININESS_BIT  0x200
+#define FRONT_INDEXES_BIT   0x400
+#define BACK_INDEXES_BIT    0x800
+
+#define FRONT_MATERIAL_BITS	(FRONT_EMISSION_BIT | FRONT_AMBIENT_BIT | \
+				 FRONT_DIFFUSE_BIT | FRONT_SPECULAR_BIT | \
+				 FRONT_SHININESS_BIT | FRONT_INDEXES_BIT)
+
+#define BACK_MATERIAL_BITS	(BACK_EMISSION_BIT | BACK_AMBIENT_BIT | \
+				 BACK_DIFFUSE_BIT | BACK_SPECULAR_BIT | \
+				 BACK_SHININESS_BIT | BACK_INDEXES_BIT)
+
+#define ALL_MATERIAL_BITS	(FRONT_MATERIAL_BITS | BACK_MATERIAL_BITS)
+
+
+
 /*
  * Specular exponent and material shininess lookup table sizes:
  */
@@ -434,7 +534,7 @@ struct gl_colorbuffer_attrib {
 	GLboolean AlphaEnabled;		/* Alpha test enabled flag */
 	GLenum AlphaFunc;		/* Alpha test function */
 	GLfloat AlphaRef;		/* Alpha reference value */
-	GLint AlphaRefInt;		/* AlphaRef scaled to an integer */
+	GLubyte AlphaRefUbyte;		/* AlphaRef scaled to an integer */
 
 	/* blending */
 	GLboolean BlendEnabled;		/* Blending enabled flag */
@@ -445,7 +545,8 @@ struct gl_colorbuffer_attrib {
 
 	/* logic op */
 	GLenum LogicOp;			/* Logic operator */
-	GLboolean LogicOpEnabled;	/* Logic op enabled flag */
+	GLboolean IndexLogicOpEnabled;	/* Color index logic op enabled flag */
+	GLboolean ColorLogicOpEnabled;	/* RGBA logic op enabled flag */
 	GLboolean SWLogicOpEnabled;	/* Do logic ops in software? */
 
 	GLboolean DitherFlag;		/* Dither enable flag */
@@ -489,7 +590,8 @@ struct gl_enable_attrib {
 	GLboolean Lighting;
 	GLboolean LineSmooth;
 	GLboolean LineStipple;
-	GLboolean LogicOp;
+	GLboolean IndexLogicOp;
+	GLboolean ColorLogicOp;
 	GLboolean Map1Color4;
 	GLboolean Map1Index;
 	GLboolean Map1Normal;
@@ -574,19 +676,20 @@ struct gl_hint_attrib {
 
 
 struct gl_light_attrib {
-	struct gl_light Light[MAX_LIGHTS];	/* Array of lights */
-	struct gl_lightmodel Model;		/* Lighting model */
-	struct gl_material Material[2];		/* Material 0=front, 1=back */
-	GLboolean Enabled;			/* Lighting enabled flag */
-	GLenum ShadeModel;			/* GL_FLAT or GL_SMOOTH */
-	GLenum ColorMaterialFace;		/* see glColorMaterial */
-	GLenum ColorMaterialMode;		/* see glColorMaterial */
-	GLboolean ColorMaterialEnabled;		/* see glColorMaterial */
+   struct gl_light Light[MAX_LIGHTS];	/* Array of lights */
+   struct gl_lightmodel Model;		/* Lighting model */
+   struct gl_material Material[2];	/* Material 0=front, 1=back */
+   GLboolean Enabled;			/* Lighting enabled flag */
+   GLenum ShadeModel;			/* GL_FLAT or GL_SMOOTH */
+   GLenum ColorMaterialFace;		/* GL_FRONT, BACK or FRONT_AND_BACK */
+   GLenum ColorMaterialMode;		/* GL_AMBIENT, GL_DIFFUSE, etc */
+   GLuint ColorMaterialBitmask;		/* bitmask formed from Face and Mode */
+   GLboolean ColorMaterialEnabled;
 
-	/* Derived for optimizations: */
-	struct gl_light *FirstEnabled;	/* Ptr to 1st enabled light */
-	GLboolean Fast;			/* Use fast shader? */
-	GLfloat BaseColor[4];
+   /* Derived for optimizations: */
+   struct gl_light *FirstEnabled;	/* Ptr to 1st enabled light */
+   GLboolean Fast;			/* Use fast shader? */
+   GLfloat BaseColor[4];
 };
 
 
@@ -659,7 +762,6 @@ struct gl_polygon_attrib {
 	GLboolean StippleFlag;	/* True if GL_POLYGON_STIPPLE is enabled */
         GLfloat OffsetFactor;	/* Polygon offset factor */
         GLfloat OffsetUnits;	/* Polygon offset units */
-        GLfloat OffsetBias;	/* = OffsetUnits / DEPTH_SCALE */
         GLboolean OffsetPoint;	/* Offset in GL_POINT mode? */
         GLboolean OffsetLine;	/* Offset in GL_LINE mode? */
         GLboolean OffsetFill;	/* Offset in GL_FILL mode? */
@@ -692,9 +794,13 @@ struct gl_stencil_attrib {
 #define S_BIT 4
 #define T_BIT 8
 
+#define TEXTURE_1D 1
+#define TEXTURE_2D 2
+#define TEXTURE_3D 4
+
 
 struct gl_texture_attrib {
-	GLuint Enabled;			/* Bitwise-OR of 1 and 2 */
+	GLuint Enabled;			/* Bitwise-OR of TEXTURE_XD values */
 	GLenum EnvMode;			/* GL_MODULATE, GL_DECAL, GL_BLEND */
 	GLfloat EnvColor[4];
 	GLuint TexGenEnabled;		/* Bitwise-OR of [QRST]_BIT values */
@@ -712,9 +818,11 @@ struct gl_texture_attrib {
 	GLfloat EyePlaneQ[4];
 	struct gl_texture_object *Current1D;
 	struct gl_texture_object *Current2D;
+	struct gl_texture_object *Current3D;
 #ifdef GL_VERSION_1_1
 	struct gl_texture_object *Proxy1D;
 	struct gl_texture_object *Proxy2D;
+	struct gl_texture_object *Proxy3D;
 #endif
 };
 
@@ -754,6 +862,8 @@ struct gl_pixelstore_attrib {
 	GLint RowLength;
 	GLint SkipPixels;
 	GLint SkipRows;
+	GLint ImageHeight;     /* for GL_EXT_texture3D */
+	GLint SkipImages;      /* for GL_EXT_texture3D */
 	GLboolean SwapBytes;
 	GLboolean LsbFirst;
 };
@@ -884,11 +994,12 @@ struct gl_evaluators {
 struct gl_texture_object {
 	GLint RefCount;			/* reference count */
 	GLuint Name;			/* an unsigned integer */
-	GLuint Dimensions;		/* 1 or 2 */
+	GLuint Dimensions;		/* 1 or 2 or 3 */
 	GLfloat Priority;		/* in [0,1] */
 	GLint BorderColor[4];		/* as integers in [0,255] */
 	GLenum WrapS;			/* GL_CLAMP or GL_REPEAT */
 	GLenum WrapT;			/* GL_CLAMP or GL_REPEAT */
+	GLenum WrapR;			/* GL_CLAMP or GL_REPEAT */
 	GLenum MinFilter;		/* minification filter */
 	GLenum MagFilter;		/* magnification filter */
 	struct gl_texture_image *Image[MAX_TEXTURE_LEVELS];
@@ -897,6 +1008,12 @@ struct gl_texture_object {
 };
 
 
+/*
+ * Display list building block:
+ */
+union node;
+
+
 /*
  * State possibly shared by multiple contexts:
  */
@@ -1014,7 +1131,7 @@ struct gl_frame_buffer {
 
 
 /*
- * Bitmaps to indicate what state has to be updated (CC.NewState)
+ * Bits to indicate what state has to be updated (CC.NewState)
  */
 #define NEW_ALL		0x7
 #define NEW_LIGHTING	0x1
@@ -1022,8 +1139,6 @@ struct gl_frame_buffer {
 #define NEW_TEXTURING	0x4
 
 
-
-
 /*
  * The library context: 
  */
@@ -1117,9 +1232,9 @@ struct gl_context {
 	GLuint RasterMask;	/* OR of rasterization flags */
 	GLuint LightTwoSide;	/* Compute two-sided lighting? */
 	GLboolean DirectTriangles;/* Directly call (*ctx->TriangleFunc) ? */
-	GLint PolygonZoffset;	/* Z offset for GL_FILL polygons */
-	GLint LineZoffset;	/* Z offset for GL_LINE polygons */
-	GLint PointZoffset;	/* Z offset for GL_POINT polygons */
+	GLfloat PolygonZoffset;	/* Z offset for GL_FILL polygons */
+	GLfloat LineZoffset;	/* Z offset for GL_LINE polygons */
+	GLfloat PointZoffset;	/* Z offset for GL_POINT polygons */
 	GLboolean NeedNormals;	 /* Are vertex normal vectors needed? */
         GLboolean FastDrawPixels;/* Use optimized glDrawPixels? */
         GLboolean MutablePixels; /* Can rasterization change pixel's color? */
@@ -1160,6 +1275,9 @@ struct gl_context {
         GLuint SwapCount;	/* number of swap-buffer calls */
         GLdouble SwapTime;	/* seconds spent in swap-buffers */
 #endif
+
+        /* For debugging/development only */
+        GLboolean NoRaster;
 };
 
 
